"""Functions to fetch and process historical race and horse data from HKJC
"""
from __future__ import annotations

import requests

# TODO read and process all races from start date to end date

# TODO query all basic info and race history for a specific horse

# TODO classify running style & draw to determine blocking probability

# 